import asyncio
import logging
import signal
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager
from dataclasses import asdict
from functools import partial
from logging.handlers import QueueHandler
from multiprocessing import Queue
from typing import Optional

import anyio
import structlog
import uvloop

from fennel.exceptions import Completed, UnknownTask
from fennel.job import Job
from fennel.utils import base64uuid, duration
from fennel.worker.broker import Broker

logger = structlog.get_logger("fennel.worker")

EXIT_SIGNAL = "signal"
EXIT_COMPLETE = "complete"


class Executor:
    def __init__(self, app):
        """
        The `Executor` is responsible for reading jobs from the Redis queue and
        executing them.

        Heartbeats are sent from the executor periodically (controlled by
        :attr:`fennel.settings.Settings.heartbeat_interval`). If they are missing for
        more than :attr:`fennel.settings.Settings.heartbeat_timeout` seconds, the
        executor will be assumed dead and all of its pending messages will be reinserted
        to the stream by another worker's maintenance function.

        Parameters
        ----------
        app : fennel.App
            The application instance for which to start an `Executor`.
        """
        self.app = app
        self.settings = app.settings
        self.executor_id: str = base64uuid()
        self.broker: Optional[Broker] = None  # Set on .start()
        self.must_stop: bool = False
        self.running: int = 0
        self.done: anyio.abc.Event = None

    def start(self, exit: str = EXIT_SIGNAL, queue: Queue = None) -> None:
        """
        Begin the main executor loop.

        Parameters
        ----------
        exit : str
            The exit strategy. `EXIT_SIGNAL` is used when the worker should only stop on
            receipt of a interrupt or termination signal. `EXIT_COMPLETE` is used in tests
            to exit when all tasks from the queue have completed.
        queue : multiprocessing.Queue
            A `QueueHandler` will be used to send logs to this queue to avoid
            interleaving from multiple processes.

        Notes
        -----
        Intended to run via :func:`fennel.worker.worker.start` which will supervise
        multiple `Executor` processes.

        `signal.SIGINT` and `signal.SIGTERM` are handled by gracefully shutting down,
        which means giving the executor processes a chance to finish their current
        tasks.
        """
        if queue:
            # To prevent interleaved logging.
            logging.getLogger("fennel.worker").addHandler(QueueHandler(queue))

        uvloop.install()
        anyio.run(self._start, exit)

    async def _start(self, exit) -> None:
        self.broker = await Broker.for_app(self.app)
        self.done = anyio.create_event()
        try:
            logger.warning("executor-started", executor=self.executor_id)
            await self._supervise(exit)
        except Completed:
            pass
        finally:
            logger.warning("executor-stopped", executor=self.executor_id)

    async def _supervise(self, exit):
        async with anyio.open_signal_receiver(signal.SIGINT, signal.SIGTERM) as signals:
            async with anyio.create_task_group() as tg:
                await tg.spawn(self._heartbeat)
                await tg.spawn(self._scheduler)
                await tg.spawn(self._maintenance)

                for i in range(self.settings.concurrency):
                    await tg.spawn(self._loop, f"{self.executor_id}:{i}", exit)

                async for signum in signals:
                    logger.critical("shutting-down")
                    async with anyio.move_on_after(self.settings.grace_period) as scope:
                        self.must_stop = True
                        await self.done.wait()

                    if scope.cancel_called:
                        logger.warning("grace-period-exceeded")

                    await tg.cancel_scope.cancel()
                    return

    @asynccontextmanager
    async def is_running(self):
        self.running += 1
        try:
            yield
        finally:
            self.running -= 1
            if self.running == 0:
                await self.done.set()

    async def _loop(self, consumer_id: str, exit: str) -> None:
        """
        The main consumer loop: read data from the stream and handle processing.
        """
        async with self.is_running():
            while not self.must_stop:
                await anyio.sleep(0)  # Check for cancellation.

                results = await self.broker.read(
                    consumer=consumer_id,
                    count=self.settings.prefetch_count,
                    recover=False,
                    timeout=self.settings.read_timeout,
                )

                if exit == EXIT_COMPLETE and not results:
                    raise Completed

                for xid, fields in results:
                    job = await self.broker.executing(fields["uuid"])
                    await self._handle(xid, job, consumer_id)

    async def _handle(self, xid: str, job: Job, consumer_id: str) -> None:
        """
        Handle the given job. All jobs are acknowledged and deleted after processing.
        Depending on the outcome, they will also be scheduled for reprocessing, their
        result will be stored, or they will be put in the dead-letter queue.
        """
        with duration(logger, "executing", xid=xid, job=asdict(job), con=consumer_id):
            job = await self._execute(job)
            job = job.increment()

            if job.exception:
                if job.tries <= job.max_retries:
                    await self.broker.ack_and_schedule(xid, job)
                    logger.debug("ack-schedule", xid=xid, job=asdict(job))
                else:
                    await self.broker.ack_and_dead(xid, job)
                    logger.debug("ack-dead", xid=xid, job=asdict(job))
            else:
                if not self.settings.results_enabled:
                    await self.broker.ack(xid, job)
                    logger.debug("ack", xid=xid, job=asdict(job))
                else:
                    await self.broker.ack_and_store(xid, job)
                    logger.debug("ack-store", xid=xid, job=asdict(job))

    async def _execute(self, job: Job) -> Job:
        """
        Attempt execution of a task. If the task is a coroutine function, then await it
        in our current event loop, otherwise run the task in a thread pool executor so
        as not to block the loop.
        """
        try:
            f = self.app.task_map[job.task]
            args, kwargs = job.args, job.kwargs
        except KeyError:
            raise UnknownTask(f"Could not find function: {job.task}")

        try:
            if asyncio.iscoroutinefunction(f):
                val = await f(*args, **kwargs)
            else:
                with ThreadPoolExecutor(max_workers=1) as pool:
                    loop = asyncio.get_running_loop()
                    val = await loop.run_in_executor(pool, partial(f, *args, **kwargs))
        except BaseException as e:
            logger.exception("execution-failed", job=asdict(job))
            return job.replace(
                exception={
                    "original_type": type(e).__name__,
                    "original_args": list(e.args),
                },
                return_value=None,
            )
        else:
            return job.replace(exception={}, return_value=val)

    async def _heartbeat(self) -> None:
        """
        Publish a heartbeat for this executor. If heartbeats are missing for more than
        settings.heartbeat_timeout seconds, the executor will be considered dead and all
        its consumers' messages will be deleted and added to the stream for
        reprocessing.
        """
        while True:
            await self.broker.heartbeat(self.executor_id)
            logger.info("heartbeat", executor=self.executor_id)
            await anyio.sleep(self.settings.heartbeat_interval)

    async def _scheduler(self) -> None:
        """
        Find any jobs whose ETA has passed and add them to the stream. Jobs are stored
        in a sorted set by their expected time of arrival and we find tasks whose value
        is less than 'now'.
        """
        while True:
            results = await self.broker.process_schedule()
            logger.info("poll-schedule", executor=self.executor_id, results=results)
            await anyio.sleep(self.settings.schedule_interval)

    async def _maintenance(self) -> None:
        """
        The maintenance script performs the following:
            1. Find dead consumers (their executor heartbeats are missing for greater
            than settings.heartbeat_timeout).
            2. Delete their pending messages and put them back in the stream for other
            consumers to process.
            3. Delete the dead consumers (and the executor's last heartbeat).
        """
        while True:
            results = await self.broker.maintenance(self.settings.heartbeat_timeout)
            logger.info("maintenance", executor=self.executor_id, results=results)
            await anyio.sleep(self.settings.maintenance_interval)



from pathlib import Path
from typing import Any, Callable, Dict, List

import aredis
import structlog

from fennel import status
from fennel.app import App
from fennel.job import Job
from fennel.keys import Keys
from fennel.utils import get_aredis, now

logger = structlog.get_logger("fennel.worker")


class Broker:
    def __init__(
        self,
        client: aredis.StrictRedis,
        keys: Keys,
        results_ttl: int,
        retry_backoff: Callable,
    ):
        """
        The `Broker` is responsible for managing communication with Redis.

        Instances are expected to be used in an event loop context and so should be
        created using :func:`fennel.worker.broker.Broker.for_app`

        Parameters
        ----------
        client : aioredis.Redis
            Used to communicate with Redis.
        keys : fennel.keys.Keys
            A collection of Redis keys configured by the app.
        results_ttl : int
            How many seconds to store task return values in Redis.
        retry_backoff : Callable
            A function of two parameters (retries: int, jitter: bool) used to set the
            time for failed tasks to wait before they are reprocessed.
        """
        self.client = client
        self.keys = keys
        self.results_ttl = results_ttl
        self.retry_backoff = retry_backoff
        self.scripts: Dict[str, Callable] = {}

    @classmethod
    async def for_app(cls, app: App) -> "Broker":
        """
        Construct the broker in an async context using this method. Actions:
            1. Create the consumer group if it doesn't already exist.
            2. Load Lua scripts.
        """
        broker = cls(
            client=get_aredis(app),
            keys=app.keys,
            results_ttl=app.settings.results_ttl,
            retry_backoff=app.settings.retry_backoff,
        )
        await broker.maybe_create_group()
        await broker.register_scripts()
        return broker

    async def read(
        self,
        consumer: str,
        count: int,
        timeout: int = 4000,
        recover: bool = False,
    ) -> List:
        """
        Read `count` jobs from the stream.
        """
        results = await self.client.xreadgroup(
            group=self.keys.group,
            consumer_id=consumer,
            count=count,
            block=timeout,
            **{self.keys.queue: "0" if recover else ">"},
        )
        return [(xid, fields) for _, result in results.items() for xid, fields in result]

    async def executing(self, uuid: str) -> Job:
        """
        Set the status entry for the given uuid to status.EXECUTING
        and return the associated Job.
        """
        async with await self.client.pipeline() as pipe:
            key = self.keys.status_prefix + f":{uuid}"
            await pipe.hmset(key, {"status": status.EXECUTING})
            await pipe.hgetall(key)
            _, fields = await pipe.execute()

        return Job.deserialise(fields)

    async def _ack(self, pipe, xid):
        key = self.keys.queue
        await pipe.xack(key, self.keys.group, xid)
        await pipe.xdel(key, xid)

    async def _store(self, pipe, job):
        key = self.keys.result(job)
        await pipe.delete(key)
        await pipe.lpush(key, job.result)
        await pipe.expire(key, self.results_ttl)

    async def _status(self, pipe, job, ttl=None):
        key = self.keys.status(job)
        await pipe.hmset(key, job.serialise())
        if ttl:
            await pipe.expire(key, ttl)

    async def _schedule(self, pipe, job, eta):
        await pipe.zadd(self.keys.schedule, eta, job.uuid)

    async def _dead(self, pipe, job):
        await pipe.xadd(self.keys.dead, {"uuid": job.uuid})

    async def ack(self, xid: str, job: Job) -> List:
        """
        Acknowledge receipt of the ID:
            1. Remove it from the consumer's PEL.
            2. Delete the message from the stream.
            3. Set the status entry for the job to status.SUCCESS.
            4. Set expiry for the status entry.
        """
        job = job.replace(status=status.SUCCESS)

        async with await self.client.pipeline() as pipe:
            await self._ack(pipe, xid)
            await self._status(pipe, job, ttl=self.results_ttl)
            return await pipe.execute()

    async def ack_and_store(self, xid: str, job: Job) -> List:
        """
        Acknowledge receipt of the ID and store the result:
            1. Remove it from the consumer's PEL.
            2. Delete the message from the stream.
            3. Delete any existing results (just in case it already exists).
            4. Store the result in a list so that clients can wait via BRPOPLPUSH.
            5. Set expiry for the result.
            6. Set the status entry for the job to status.SUCCESS.
            7. Set expiry for the status entry.
        """
        job = job.replace(status=status.SUCCESS)

        async with await self.client.pipeline() as pipe:
            await self._ack(pipe, xid)
            await self._store(pipe, job)
            await self._status(pipe, job, ttl=self.results_ttl)
            return await pipe.execute()

    async def ack_and_schedule(self, xid: str, job: Job) -> List:
        """
        Acknowledge receipt of the ID and schedule the job for reprocessing:
            1. Remove it from the consumer's PEL.
            2. Delete the message from the stream.
            3. Add the job to the schedule sorted set so that consumers can poll it.
            4. Set the status entry for the job to status.RETRY.
        """
        job = job.replace(status=status.RETRY)
        eta = now() + int(self.retry_backoff(job.tries))

        async with await self.client.pipeline() as pipe:
            await self._ack(pipe, xid)
            await self._schedule(pipe, job, eta)
            await self._status(pipe, job, ttl=None)
            return await pipe.execute()

    async def ack_and_dead(self, xid: str, job: Job) -> List:
        """
        Acknowledge receipt of the ID and add the job to the dead-letter queue:
            1. Remove it from the consumer's PEL.
            2. Delete the message from the stream.
            3. Add the message to the DLQ.
            4. Delete any existing results (just in case it already exists).
            5. Store the result in a list so that clients can wait via BRPOPLPUSH.
            6. Set expiry for the result.
            7. Set the status entry for the job to status.DEAD.
        """
        job = job.replace(status=status.DEAD)

        async with await self.client.pipeline() as pipe:
            await self._ack(pipe, xid)
            await self._dead(pipe, job)
            await self._store(pipe, job)
            await self._status(pipe, job, ttl=None)
            return await pipe.execute()

    async def process_schedule(self) -> List:
        """
        Retrieve any jobs whose ETA has passed and add them to the stream.
        """
        fn = self.scripts["schedule"]
        return await fn.execute(
            keys=[self.keys.schedule, self.keys.queue],
            args=[now()],
        )

    async def heartbeat(self, executor_id: str) -> List:
        """
        Publish a heartbeat timestamp for the given worker.
        """
        return await self.client.hset(self.keys.heartbeats, executor_id, now())

    async def maintenance(self, threshold: int) -> List:
        """
        Execute the maintenance script:
            1. Find dead consumers (their worker heartbeats are missing for greater than
            settings.heartbeat_timeout).
            2. Delete their pending messages and put them back in the stream for other
            consumers to process.
            3. Delete the dead consumers (and the worker's last heartbeat).
        """
        fn = self.scripts["maintenance"]
        return await fn.execute(
            keys=[self.keys.queue, self.keys.heartbeats],
            args=[self.keys.group, now(), threshold],
        )

    async def create_group(self) -> Any:
        """
        Create the consumer group and the streams.
        """
        for key in [self.keys.queue, self.keys.dead]:
            cmd = ["XGROUP", "CREATE", key, self.keys.group, "0", "MKSTREAM"]
            await self.client.execute_command(*cmd)

    async def maybe_create_group(self) -> Any:
        """
        Create the consumer group (and the streams) if they don't exist.
        """
        try:
            return await self.create_group()
            logger.debug("group-created", group=self.keys.group)
        except aredis.ResponseError as e:
            if str(e).startswith("BUSYGROUP"):
                logger.debug("group-exists", group=self.keys.group)
            else:
                raise

    async def register_scripts(self) -> None:
        """
        Load the Lua scripts into Redis.
        """
        for script in (Path(__file__).parent / "lua").glob("*.lua"):
            self.scripts[script.stem] = self.client.register_script(script.read_text())



from typing import Optional

import click

from fennel import client
from fennel.client.state import get_job, get_state
from fennel.exceptions import JobNotFound
from fennel.models import render
from fennel.utils import get_object
from fennel.worker import start


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.option(
    "-a",
    "--app",
    "application",
    required=True,
    help="""
    A colon-separated string identifying the :class:`fennel.App` instance for which to
    run a worker.

    If a file ``foo.py`` exists at the current working directory with the following
    contents:

    >>> from fennel import App
    >>>
    >>> app = App(name="myapp", redis_url="redis://127.0.0.1:6379")
    >>>
    >>> @app.task
    >>> def f():
    >>>     pass

    Then pass ``foo:app`` as the `app` option: ``$ fennel worker --app=foo:app``
""",
)
@click.option(
    "-p",
    "--processes",
    type=int,
    default=None,
    help="""
    How many executor processes to run in each worker. Default
    ``multiprocessing.cpu_count()``
""",
)
@click.option(
    "-c",
    "--concurrency",
    type=int,
    default=None,
    help="""
    How many concurrent consumers to run (we make at least this many Redis
    connections) in each executor process. The default, 8, can handle 160 req/s in
    a single worker process if each task is IO-bound and lasts on average 50ms. If
    you have long running CPU-bound tasks, you will want to run multiple executor
    processes. Default ``8``
""",
)
def worker(application: str, processes: int, concurrency: int) -> None:
    """
    Run the worker.
    """
    app = get_object(application)
    if processes:
        app.settings.processes = processes
    if concurrency:
        app.settings.concurrency = concurrency
    start(app)


@cli.command()
@click.option("-a", "--app", "application", required=True)
@click.argument("action", type=click.Choice(["read", "replay", "purge"]))
def dlq(application: str, action: str) -> None:
    """
    Interact with the dead-letter queue. Choices for the `action` argument:

    \b
    * read - Print all tasks from the dead-letter queue to stdout.
    * replay - Move all tasks from the dead-letter queue back to the main task queue for reprocessing.
    * purge - Remove all tasks from the dead-letter queue forever.
    """
    if action == "read":
        click.echo(render(client.read_dead(get_object(application))))
    elif action == "replay":
        click.echo(render(client.replay_dead(get_object(application))))
    elif action == "purge":
        click.echo(render(client.purge_dead(get_object(application))))


@cli.command()
@click.option("-a", "--app", "application", required=True)
@click.option("-u", "--uuid", required=True)
def task(application: str, uuid: Optional[str]) -> None:
    """
    Print a JSON-encoded summary of job information.
    """
    app = get_object(application)
    try:
        click.echo(render(get_job(app, uuid)))
    except JobNotFound:
        click.echo(render({"error": f"Task with uuid={uuid} not found"}))


@cli.command()
@click.option("-a", "--app", "application", required=True)
def info(application: str) -> None:
    """
    Print a JSON-encoded summary of application state.
    """
    click.echo(render(get_state(get_object(application))))



import asyncio
from typing import Iterable

from fennel.client.results import AsyncResult


def gather(results: Iterable[AsyncResult], task_timeout=10, return_exceptions=True):
    """
    Multi-result version of .get() -- wait for all tasks to complete and return all of
    their results in order.

    Has the same semantics as `asyncio.gather`.
    """
    async def _gather():
        async def _get(result, timeout):
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, result.get, timeout)

        aws = [_get(r, task_timeout) for r in results]
        return await asyncio.gather(*aws, return_exceptions=return_exceptions)

    return asyncio.run(_gather())


def wait(results: Iterable[AsyncResult], timeout: int, return_when="ALL_COMPLETED"):
    """
    Wait for all tasks to complete and return two sets of Futures (done, pending).

    Has the same semantics as `asyncio.wait`.
    """
    async def _wait():
        async def _get(result):
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, result.get)

        aws = [_get(r) for r in results]
        return await asyncio.wait(aws, timeout=timeout, return_when=return_when)

    return asyncio.run(_wait())



import base64
import importlib
import multiprocessing as mp
import random
import time
import uuid
from contextlib import contextmanager
from typing import Any, Generator

import aredis
import redis

EMPTY = object()  # A unique value used to distinguish emptiness from None


def backoff(retries: int, jitter: bool = True) -> int:
    """
    Compute duration (seconds) to wait before retrying using exponential backoff with
    jitter based on the number of retries a message has already experienced.

    The minimum returned value is 1s
    The maximum returned value is 604800s (7 days)

    With max_retries=9, you will have roughly 30 days to fix and redeploy the the task
    code.

    Parameters
    ----------
    retries : int
        How many retries have already been attemped.
    jitter : bool
        Whether to add random noise to the return value (recommended).

    Notes
    -----
    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    """
    x = 6**(retries + 1)
    if jitter:
        x = random.randrange(x // 3, x * 2)
    return min(604_800, x)


def base64uuid() -> str:
    return base64.urlsafe_b64encode(uuid.uuid4().bytes).rstrip(b"=").decode("ascii")


def now() -> float:
    return time.time()


@contextmanager
def duration(logger, event: str, **extra: Any) -> Generator:
    logger.info(f"started-{event}", **extra)
    start = time.perf_counter()
    yield
    duration = time.perf_counter() - start
    logger.info(f"ended-{event}", duration=duration, **extra)


def get_object(name: str) -> Any:
    module_name, instance_name = name.split(":", 1)
    module = importlib.import_module(module_name)
    return getattr(module, instance_name)


def get_redis(app):
    return redis.Redis.from_url(
        app.settings.redis_url,
        decode_responses=True,
    )


def get_aredis(app):
    return aredis.StrictRedis.from_url(
        app.settings.redis_url,
        decode_responses=True,
    )


def get_mp_context():
    # The default changed to 'spawn' in 3.8 on macOS due to https://bugs.python.org/issue33725
    return mp.get_context("fork")



import json

from fennel.exceptions import Timeout
from fennel.job import Job
from fennel.status import SENT


def send(app, job, xid=None):
    """
    Send the job to the main task queue with the status SENT.
    """
    job = job.replace(status=SENT)
    with app.client.pipeline() as pipe:
        pipe.hset(app.keys.status(job), mapping=job.serialise())
        pipe.xadd(app.keys.queue, fields={"uuid": job.uuid}, id=xid or "*")
        return pipe.execute()


def result(app, job, timeout):
    """
    Retrieve the result of executing the given `job`, blocking for `timeout` seconds
    before raising Timeout.
    """
    key = app.keys.result(job)
    value = app.client.brpoplpush(key, key, timeout)

    if value is None:
        raise Timeout
    return json.loads(value)


def read_dead(app, batchsize=100):
    """
    Iterate over the dead-letter queue and return all job data.
    """
    return list(_iter_dlq(app, batchsize=batchsize))


def replay_dead(app, filter=lambda job: True, batchsize=100):
    """
    Iterate over the dead-letter queue and replay any jobs for which filter(job)
    evaluates to True. The default is to replay all jobs.
    """
    jobs = list(_iter_dlq(app, batchsize=batchsize))
    results = []
    with app.client.pipeline() as pipe:
        for xid, job in jobs:
            if filter(job):
                pipe.hset(app.keys.status(job), "status", SENT)
                pipe.xadd(app.keys.queue, {"uuid": job.uuid})
                pipe.xdel(app.keys.dead, xid)
                results.append((xid, job))
        pipe.execute()
    return results


def purge_dead(app, filter=lambda job: True, batchsize=100):
    """
    Iterate over the dead-letter queue and delete any jobs for which filter(job)
    evaluates to True. The default is to delete all jobs.
    """
    jobs = list(_iter_dlq(app, batchsize=batchsize))
    results = []
    with app.client.pipeline() as pipe:
        for xid, job in jobs:
            if filter(job):
                pipe.delete(app.keys.status(job))
                pipe.xdel(app.keys.dead, xid)
                results.append((xid, job))
        pipe.execute()
    return results


def _iter_dlq(app, batchsize=100):
    xid = "0"

    while True:
        response = app.client.xread({app.keys.dead: xid}, count=batchsize)
        if not response:
            return

        for stream, messages in response:
            assert stream == app.keys.dead

            if not messages:
                return
            else:
                xid = messages[-1][0]

            with app.client.pipeline() as pipe:
                for xid, fields in messages:
                    pipe.hgetall(app.keys.status_prefix + f":{fields['uuid']}")
                response = pipe.execute()

            yield from zip(
                [xid for xid, _ in messages],
                [Job.deserialise(fields) for fields in response],
            )



import json
from dataclasses import asdict, field, replace
from typing import Any, Dict, List

from pydantic import validator
from pydantic.dataclasses import dataclass

from fennel.status import UNKNOWN
from fennel.utils import base64uuid


@dataclass
class Job:
    """
    The internal representation of a job.

    Parameters
    ----------
    task : str
        The name of the task. By default will use
        ``f"{func.__module__}.{func.__qualname__}"``, where `func` is the Python
        callable.
    args : List
        The job's args.
    kwargs : Dict
        The job's kwargs.
    tries : int
        The number of attempted executions.
    max_retries : int
        The maximum number of retries to attempt after failure.
    exception : Dict
        Exception information for the latest failure, contains
        'original_type' (str, e.g. 'ValueError') and
        'original_args' (List, e.g. ['Not found']).
    return_value : Any
        The return value of the Python callable when execution succeeds.
    status : str
        One of :mod:`fennel.status`, the current lifecycle stage.
    uuid : str
        Base64-encoded unique identifier.
    """
    task: str
    args: List
    kwargs: Dict
    tries: int = 0
    max_retries: int = 9
    exception: Dict = field(default_factory=dict)
    return_value: Any = None
    status: str = UNKNOWN
    uuid: str = field(default_factory=base64uuid)

    @validator("args", "kwargs", "exception", "return_value", pre=True, whole=True)
    def json_decode(cls, v):
        if isinstance(v, str):
            try:
                return json.loads(v)
            except ValueError:
                pass
        return v

    def increment(self) -> "Job":
        return replace(self, tries=self.tries + 1)

    def replace(self, **kwargs):
        return replace(self, **kwargs)

    def to_json(self) -> str:
        return json.dumps(asdict(self))

    @classmethod
    def from_json(cls, s: str) -> "Job":
        return cls(**json.loads(s))

    def serialise(self) -> Dict:
        d = asdict(self)
        d["args"] = json.dumps(d["args"])
        d["kwargs"] = json.dumps(d["kwargs"])
        d["exception"] = json.dumps(d["exception"])
        d["return_value"] = json.dumps(d["return_value"])
        return d

    @classmethod
    def deserialise(cls, fields: Dict) -> "Job":
        return cls(**fields)

    @property
    def result(self) -> str:
        return json.dumps({
            "return_value": self.return_value,
            "exception": self.exception,
        })



from typing import List


class FennelException(Exception):
    pass


class TaskFailed(FennelException):
    def __init__(self, original_type: str, original_args: List):
        """
        This exception is returned by worker processes which experienced an exception
        when executing a task.

        Parameters
        ----------
        original_type : str
            The name of the original exception, e.g. ``'ValueError'``.
        original_args : List
            The arguments given to the original exception, e.g. ``['Not found']``

        Examples
        --------
        >>> @app.task(retries=0)
        >>> async def foo(n):
        ...     raise Exception("baz")
        ...
        >>> x = await foo.delay(3)
        >>> try:
        ...     result = await x.get()
        >>> except TaskFailed as e:
        ...     assert e.original_type == "Exception"
        ...     assert e.original_args == ["baz"]
        """
        self.original_type = original_type
        self.original_args = original_args


class ResultsDisabled(FennelException):
    """
    Raised when ``results_enabled=False`` and code attempts to access a tasks result via
    ``.get()``.
    """


class UnknownTask(FennelException):
    """
    Raised by a worker process if it is unable to find a Python function corresponding
    to the task it has read from the queue.
    """


class Timeout(FennelException):
    """
    Raised by client code when a given timeout is exceeded when waiting for results to arrive.
    """


class JobNotFound(FennelException):
    """
    Raised by client code when attempting to retrieve job information that cannot be
    found in Redis.
    """


class Chaos(FennelException):
    """
    Used in tests to ensure failures are handled properly.
    """


class Completed(FennelException):
    """
    Used internally to shutdown an Executor if the exit condition is completing all
    tasks.
    """



import logging
import multiprocessing as mp
import signal
import time
from logging.handlers import QueueListener

import structlog

from fennel.utils import get_mp_context
from fennel.worker.discovery import autodiscover
from fennel.worker.executor import EXIT_SIGNAL, Executor

logger = structlog.get_logger("fennel.worker")


def start(app, exit=EXIT_SIGNAL):
    """
    The main entrypoint for the worker.

    The worker will create and monitor `N` :class:`fennel.worker.Executor` processes. Each
    `Executor` will spawn `M` coroutines via an asyncio event loop. `N` and `M` are
    controlled by :attr:`fennel.settings.Settings.processes` and
    :attr:`fennel.settings.Settings.concurrency` respectively.

    CPU-bound tasks benefit from multiple processes. IO-bound tasks will benefit from
    high executor concurrency.

    Parameters
    ----------
    app : fennel.App
        The application instance for which to start a background worker.
    exit : str
        The exit strategy. `EXIT_SIGNAL` is used when the worker should only stop on
        receipt of a interrupt or termination signal. `EXIT_COMPLETE` is used in tests
        to exit when all tasks from the queue have completed.

    Notes
    -----
    `signal.SIGINT` and `signal.SIGTERM` are handled by gracefully shutting down, which
    means giving the executor processes a chance to finish their current tasks.
    """
    autodiscover(app)
    running = True

    def stop(signum, frame):
        nonlocal running
        running = False

        if signum == 2:
            logger.critical("sigint")
        elif signum == 15:
            logger.critical("sigterm")

    signal.signal(signal.SIGINT, stop)
    signal.signal(signal.SIGTERM, stop)

    ctx = get_mp_context()
    queue: mp.Queue = ctx.Queue()  # To avoid interleaving executor process logs.

    processes = [
        ctx.Process(target=Executor(app).start, args=(exit, queue), daemon=True)
        for _ in range(app.settings.processes)
    ]

    for p in processes:
        p.start()

    # Start the logging listener thread after forking.
    QueueListener(queue, logging.StreamHandler()).start()

    try:
        while running and all(p.is_alive() for p in processes):
            time.sleep(0.1)
    finally:
        for p in processes:
            p.terminate()
        for p in processes:
            p.join()
        logger.critical("exit")



import json
from datetime import datetime

from fennel import models
from fennel.exceptions import JobNotFound
from fennel.job import Job


def get_status(app, uuid):
    resp = app.client.hget(app.keys.status_prefix + f":{uuid}", "status")
    if not resp:
        raise JobNotFound(uuid)
    return resp


def get_job(app, uuid):
    resp = app.client.hgetall(app.keys.status_prefix + f":{uuid}")
    if not resp:
        raise JobNotFound(uuid)
    return Job(**resp)


def count_jobs(app):
    return len(list(app.client.scan_iter(app.keys.status_prefix + ":*")))


def get_messages(app, key, n=100):
    messages = app.client.xrevrange(key, count=n)
    return [models.Message(id=id, uuid=fields["uuid"]) for id, fields in messages]


def get_info(app, key):
    info = app.client.xinfo_stream(key)
    return models.Info(**{k.replace("-", "_"): v for k, v in info.items()})


def get_groups(app, key):
    groups = app.client.xinfo_groups(key)
    return [models.Group(**{k.replace("-", "_"): v for k, v in d.items()}) for d in groups]


def get_stream(app, key, n=100):
    return models.Stream(
        key=key,
        info=get_info(app, key),
        groups=get_groups(app, key),
        messages=get_messages(app, key, n=n),
    )


def get_result(app, key):
    value = json.loads(app.client.lrange(key, 0, 1)[0])
    return models.Result(return_value=value["return_value"], exception=value["exception"])


def count_results(app):
    return len(list(app.client.scan_iter(app.keys.result_prefix + ":*")))


def get_heartbeats(app):
    beats = app.client.hgetall(app.keys.heartbeats)
    return [
        models.Heartbeat(executor_id=k, timestamp=datetime.fromtimestamp(float(v)))
        for k, v in zip(beats.keys(), beats.values())
    ]


def get_schedule(app):
    return [
        models.ScheduledJob(eta=datetime.fromtimestamp(int(v)), uuid=k)
        for k, v in app.client.zscan_iter(app.keys.schedule)
    ]


def get_state(app, messages=100):
    return models.State(
        queue=get_stream(app, app.keys.queue, n=messages),
        dead=get_stream(app, app.keys.dead, n=messages),
        heartbeats=get_heartbeats(app),
        schedule=get_schedule(app),
    )



import importlib
from pathlib import Path

import structlog

logger = structlog.get_logger("fennel.worker")


def autodiscover(app, basedir=None):
    """
    Automatically import Python modules matching the pattern in
    :attr:`fennel.settings.Settings.autodiscover`.

    Tasks are registered via decorators. When the worker process starts up, these
    decorators must be executed in order for the app to record which Python functions
    map to which task names.

    Parameters
    ----------
    app : fennel.App
        The app instance for which to discover tasks.
    basedir : str
        The directory relative to which we will search the filesystem.
    """
    pattern = app.settings.autodiscover

    if pattern:
        for filename in Path(basedir or Path('.')).glob(pattern):
            module = ".".join(filename.parts)[:-3]
            logger.debug("discovered", module=module)
            importlib.import_module(module)



import asyncio
from typing import Iterable

from fennel.client.aio.results import AsyncResult


async def gather(results: Iterable[AsyncResult], task_timeout=10, return_exceptions=True):
    """
    Multi-result version of .get() -- wait for all tasks to complete and return all of
    their results in order.

    Has the same semantics as `asyncio.gather`.
    """
    aws = [r.get(timeout=task_timeout) for r in results]
    return await asyncio.gather(*aws, return_exceptions=return_exceptions)


async def wait(results: Iterable[AsyncResult], timeout: int, return_when="ALL_COMPLETED"):
    """
    Wait for all tasks to complete and return two sets of Futures (done, pending).

    Has the same semantics as `asyncio.wait`.
    """
    aws = [r.get() for r in results]
    return await asyncio.wait(aws, timeout=timeout, return_when=return_when)



import json
from typing import Any, List

from fennel.exceptions import Timeout
from fennel.job import Job
from fennel.status import SENT


async def send(app, job, xid=None) -> List:
    """
    Send the job to the main task queue with the status SENT.
    """
    job = job.replace(status=SENT)

    async with await app.aioclient.pipeline() as pipe:
        await pipe.xadd(app.keys.queue, entry={"uuid": job.uuid}, stream_id=xid or "*")
        await pipe.hmset(app.keys.status(job), job.serialise())
        return await pipe.execute()


async def result(app, job, timeout) -> Any:
    """
    Retrieve the result of executing the given `job`, blocking for `timeout` seconds
    before raising Timeout.
    """
    key = app.keys.result(job)
    value = await app.aioclient.brpoplpush(key, key, timeout)

    if value is None:
        raise Timeout
    return json.loads(value)


async def read_dead(app, batchsize=100):
    """
    Iterate over the dead-letter queue and return all job data.
    """
    return [x async for x in _iter_dlq(app, batchsize=batchsize)]


async def replay_dead(app, filter=lambda job: True, batchsize=100):
    """
    Iterate over the dead-letter queue and replay any jobs for which filter(job)
    evaluates to True. The default is to replay all jobs.
    """
    jobs = [x async for x in _iter_dlq(app, batchsize=batchsize)]
    results = []

    async with await app.aioclient.pipeline() as pipe:
        for xid, job in jobs:
            if filter(job):
                await pipe.hset(app.keys.status(job), "status", SENT)
                await pipe.xadd(app.keys.queue, {"uuid": job.uuid})
                await pipe.xdel(app.keys.dead, xid)
                results.append((xid, job))
        await pipe.execute()
    return results


async def purge_dead(app, filter=lambda job: True, batchsize=100):
    """
    Iterate over the dead-letter queue and delete any jobs for which filter(job)
    evaluates to True. The default is to delete all jobs.
    """
    jobs = [x async for x in _iter_dlq(app, batchsize=batchsize)]
    results = []

    async with await app.aioclient.pipeline() as pipe:
        for xid, job in jobs:
            if filter(job):
                await pipe.delete(app.keys.status(job))
                await pipe.xdel(app.keys.dead, xid)
                results.append((xid, job))
        await pipe.execute()
    return results


async def _iter_dlq(app, batchsize=100):
    """
    Iterate over the dead-letter queue and replay any jobs for which filter(job)
    evaluates to True. The default is to replay all jobs.
    """
    xid = "0"

    while True:
        response = await app.aioclient.xread(
            count=batchsize,
            block=None,
            **{app.keys.dead: xid},
        )
        if not response:
            return

        for stream, messages in response.items():
            assert stream == app.keys.dead

            if not messages:
                return
            else:
                xid = messages[-1][0]
                print(xid)

            async with await app.aioclient.pipeline() as pipe:
                for xid, fields in messages:
                    await pipe.hgetall(app.keys.status_prefix + f":{fields['uuid']}")
                response = await pipe.execute()

            jobs = zip(
                [xid for xid, _ in messages],
                [Job.deserialise(fields) for fields in response],
            )

            for xid, job in jobs:
                yield (xid, job)



from dataclasses import dataclass


@dataclass
class Keys:
    queue: str
    group: str
    dead: str
    schedule: str
    heartbeats: str
    status_prefix: str
    result_prefix: str

    @classmethod
    def for_app(cls, app):
        key = lambda x: f"__fennel:{x}:{app.name}"
        return cls(
            queue=key("queue"),
            group=key("group"),
            dead=key("dead"),
            schedule=key("sched"),
            heartbeats=key("beats"),
            status_prefix=key("status"),
            result_prefix=key("result"),
        )

    def status(self, job):
        return self.status_prefix + f":{job.uuid}"

    def result(self, job):
        return self.result_prefix + f":{job.uuid}"



from functools import wraps
from typing import Any, Callable, Dict

import structlog

from fennel.client import Task as SyncTask
from fennel.client.aio import Task as AsyncTask
from fennel.keys import Keys
from fennel.logging import init_logging
from fennel.settings import Settings
from fennel.utils import EMPTY, get_aredis, get_redis

logger = structlog.get_logger("fennel.client")


class App:
    def __init__(self, name: str, **kwargs):
        """
        The app is the main abstraction provided by Fennel. Python functions are
        decorated via ``@app.task`` to enable background processing. All settings are
        configured on this object.

        Parameters
        ----------
        name : str
            Used to identify this application, e.g. to set which tasks
            a worker will execute.
        kwargs
            Any settings found in :class:`fennel.settings.Settings`

        Examples
        --------
        >>> from fennel import App
        ...
        >>> app = App(
        ...     name='myapp',
        ...     redis_url='redis://127.0.0.1',
        ...     default_retries=3,
        ...     results_enabled=True,
        ...     log_level='info',
        ...     log_format='json',
        ...     autodiscover='**/tasks.py',
        ...     interface='sync',
        ... )
        ...
        >>> @app.task(retries=1)
        >>> def foo(x):
        ...     return x
        ...
        >>> x = foo.delay(7)  # Execute in the background.
        >>> x
        AsyncResult(uuid=Tjr75jM3QDOHoLTLyrsY1g)
        >>> x.get()  # Wait for the result.
        7

        If your code is running in an asynchronous event loop (e.g. via Starlette,
        FastAPI, Quart), you will want to use the async interface instead:

        >>> import asyncio
        ...
        >>> app = App(name='foo', interface='async')
        ...
        >>> @app.task
        >>> async def bar(x)
        ...     await asyncio.sleep(x)
        ...     return x
        ...
        >>> x = await bar.delay(5)
        >>> await x.status()
        SENT
        >>> await x.get()
        5
        """
        self.name: str = name
        self.settings: Settings = Settings(**kwargs)
        self.task_map: Dict[str, Callable] = {}
        self.keys: Keys = Keys.for_app(self)
        self.client = get_redis(self)
        self.aioclient = get_aredis(self)

        if self.settings.interface == "sync":
            self.task_class = SyncTask
        else:
            self.task_class = AsyncTask

        init_logging(level=self.settings.log_level, format=self.settings.log_format)

    def task(self, func: Callable = None, *, name=None, retries=EMPTY) -> Any:
        """
        A decorator to register a function with the app to enable background processing
        via the task queue.

        The worker process (see :class:`fennel.worker.worker`) will need to discover all
        registered tasks on startup. The means all the modules containing tasks need to
        be imported. Fennel will import modules found via
        :attr:`fennel.settings.Settings.autodiscover`, which by default is
        ``'**/tasks.py'``.

        Parameters
        ---------
        func : Callable
            The decorated function.
        name : str
            The representation used to uniquely identify this task.
        retries : int
            The number of attempts at execution after a task has failed (meaning raised
            any exception).

        Examples
        --------
        Exposes an interface similar to Celery:

        >>> @app.task(retries=1)
        >>> def foo(x):
        ...     return x

        Tasks can be enqueued for processing via:

        >>> foo.delay(8)
        AsyncResult(uuid=q_jb6KaUT-G4tOAoyQ0yaA)

        The can also be called normally, bypassing the Fennel system entirely:

        >>> foo(3)
        3

        By default, tasks are 'fire-and-forget', meaning we will not wait for their
        completion. They will be executed by worker process and will be retried
        automatically on failure (using exponential backoff), so we assume tasks are
        idempotent.

        You can also wait for the result:

        >>> x = foo.delay(4)
        >>> x.status()
        SENT
        >>> x.get(timeout=10)
        4

        If instead you have many tasks and wish to wait for them to complete you can use
        the waiting primitives provided (you will want to ensure all tasks have
        retries=0, which you can set by default with an app setting):

        >>> from fennel.client import gather, wait
        >>> results = [foo.delay(x) for x in range(10)]
        >>> gathered = gather(results)  # Or:
        >>> done, pending = wait(results, timeout=2)

        If your application is running in an event loop you can elect to use the async
        interface for your fennel app (see :attr:`fennel.settings.Settings.interface`),
        which uses `aioredis` under the hood to enqueue items, retrieve results, etc, so
        you will need to await those coroutines:

        >>> app = App(name='foo', interface='async')
        >>>
        >>> @app.task
        >>> async def bar(x)
        ...     await asyncio.sleep(x)
        >>>
        >>> x = await bar.delay(1)
        >>> await x.status()
        SUCCESS
        """
        @wraps(func)
        def wrapper(f: Callable):
            nonlocal retries
            retries = retries if retries is not EMPTY else self.settings.default_retries
            assert type(retries) == int, "retries must be an integer"
            assert retries >= 0, "retries must be an integer >= 0"

            nonlocal name
            name = name or f"{f.__module__}.{f.__qualname__}"
            assert type(name) == str, "name must be a str"
            assert name not in self.task_map, f"Task already found by app: {name}"

            logger.debug("found-task", name=name)
            self.task_map[name] = f
            return self.task_class(name, f, retries=retries, app=self)

        return wrapper(func) if func else wrapper

    def __repr__(self) -> str:
        return f"App(name={self.name}, redis={self.settings.redis_url})"



from typing import Any, Dict

from fennel.client.aio.actions import result
from fennel.client.aio.state import get_status
from fennel.exceptions import ResultsDisabled, TaskFailed
from fennel.job import Job
from fennel.utils import EMPTY


class AsyncResult:
    def __init__(self, job: Job, app):
        """
        A handle for a task that is being processed by workers via the task queue.

        Conceptually similar to the AsyncResult from the mutliprocessing library.
        """
        self.job = job
        self.app = app
        self._result: Dict = None

    async def status(self):
        """
        Return the status of the task execution.

        Examples
        --------
        >>> @app.task
        >>> async def bar(x)
        ...     await asyncio.sleep(x)
        ...     return x
        ...
        >>> x = await bar.delay(5)
        >>> await x.status()
        SENT
        >>> await x.status()  # After roughly 5 seconds...
        SUCCESS
        """
        return await get_status(self.app, self.job.uuid)

    async def get(self, timeout: int = EMPTY) -> Any:
        """
        Wait for the result to become available and return it.

        Raises
        ------
        :exc:`fennel.exceptions.TaskFailed`
            If the original function raised an exception.
        :exc:`fennel.exceptions.Timeout`
            If > `timeout` seconds elapse before a result is available.

        Examples
        --------
        >>> @app.task(retries=0)
        >>> def foo(x):
        ...     return x
        ...
        >>> x = await foo.delay(7)
        >>> await x.get()  # Wait for the result.
        7

        Warning
        -------
        You must have results storage enabled
        (:attr:`fennel.settings.Settings.results_enabled`)

        If you have retries enabled, they may be rescheduled many times, so you may
        prefer to use retries=0 for tasks whose result you intend to wait for.
        """
        if not self.app.settings.results_enabled:
            raise ResultsDisabled

        if timeout is EMPTY:
            timeout = self.app.settings.task_timeout

        if self._result is None:
            self._result = await result(self.app, self.job, timeout)  # Raises Timeout

        exc, val = self._result["exception"], self._result["return_value"]

        if exc:
            raise TaskFailed(**exc)
        else:
            return val

    def __repr__(self) -> str:
        return f"AsyncResult(uuid={self.job.uuid})"



import json
from dataclasses import asdict, is_dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from pydantic.dataclasses import dataclass


def render(content) -> bytes:
    s = json.dumps(content, cls=Encoder, indent=4, separators=(",", ": "))
    return s.encode("utf-8")


class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        if is_dataclass(obj):
            return asdict(obj)
        return json.JSONEncoder.default(self, obj)


@dataclass
class Message:
    id: str
    uuid: str


@dataclass
class Info:
    length: int
    radix_tree_keys: int
    radix_tree_nodes: int
    groups: int
    last_generated_id: str
    first_entry: Optional[Tuple[str, Dict]]
    last_entry: Optional[Tuple[str, Dict]]


@dataclass
class Group:
    name: str
    consumers: int
    pending: int
    last_delivered_id: str


@dataclass
class Stream:
    key: str
    info: Info
    groups: List[Group]
    messages: List[Message]


@dataclass
class Result:
    return_value: Any
    exception: Dict


@dataclass
class Heartbeat:
    executor_id: str
    timestamp: datetime


@dataclass
class ScheduledJob:
    eta: datetime
    uuid: str


@dataclass
class State:
    queue: Optional[Stream]
    dead: Optional[Stream]
    heartbeats: List[Heartbeat]
    schedule: List[ScheduledJob]



import multiprocessing

from pydantic import BaseSettings, PyObject, validator


class Settings(BaseSettings):
    """
    Settings can be configured via environment variables or keyword arguments for the
    fennel.App instance (which take priority).

    Examples
    --------
    For environment variables, the prefix is ``FENNEL_``, for instance:

    | ``FENNEL_REDIS_URL=redis://127.0.0.1:6379``
    | ``FENNEL_DEFAULT_RETRIES=3``
    | ``FENNEL_RESULTS_ENABLED=true``

    Or via App kwargs:

    >>> from fennel import App
    ...
    >>> app = App(
    ...     name='myapp',
    ...     redis_url='redis://127.0.0.1',
    ...     default_retries=3,
    ...     results_enabled=True,
    ...     log_level='info',
    ...     log_format='json',
    ...     autodiscover='**/tasks.py',
    ...     interface='sync',
    ... )

    Parameters
    ----------
    redis_url : str
        Redis URL. Default ``'redis://127.0.0.1:6369'``
    interface : str
        Which client interface should we use -- sync or async? Default ``'sync'``
    processes : int
        How many executor processes to run in each worker. Default
        ``multiprocessing.cpu_count()``
    concurrency : int
        How many concurrent consumers to run (we make at least this many Redis
        connections) in each executor process. The default, 8, can handle 160 req/s in
        a single executor process if each task is IO-bound and lasts on average 50ms. If
        you have long running CPU-bound tasks, you will want to run multiple executor
        processes (and set heartbeat_timeout to greater than your maximum expected task
        duration). Default ``8``
    default_retries : int
        How many times to retry a task in case it raises an exception during execution.
        With 10 retries and the default :func:`fennel.utils.backoff` function, this will
        be approximately 30 days of retries. Default ``10``
    retry_backoff : Callable
        Which algorithm to use to determine the retry schedule. The default is
        exponential backoff via :func:`fennel.utils.backoff`.
    read_timeout : int
        How many milliseconds to wait for messages in the main task queue. Default
        ``4000``
    prefetch_count : int
        How many messages to read in a single call to `XREADGROUP`. Default ``1``
    heartbeat_timeout : float
        How many seconds before an executor is considered dead if heartbeats are missed.
        If you have long-running CPU-bound tasks, this value should be greater than your
        maximum expected task duration. Default ``60``
    heartbeat_interval : float
        How many seconds to sleep between heartbeats are stored for each executor process.
        Default ``6``
    schedule_interval : float
        How many seconds to sleep between polling for scheduled tasks. Default ``4``
    maintenance_interval : float
        How many seconds to sleep between running the maintenance script. Default ``8``
    task_timeout : int
        How long to wait for results to be computed when calling .get(), seconds.
        Default ``10``
    grace_period : int
        How many seconds to wait for in-flight tasks to complete before forcefully
        exiting. Default: ``30``
    restults_enabled : bool
        Whether to store results. Can be disabled if your only use-case is
        'fire-and-forget'. Default ``True``
    results_ttl : int
        How long before expiring results in seconds. Default ``3600`` (one hour).
    log_format : str
        Whether to pretty print a human-readable log ("console") or JSON ("json").
        Default ``'console'``
    log_level : str
        The minimum log level to emit. Default ``'debug'``
    autodiscover : str
        The pattern for :func:`pathlib.Path.glob` to find modules containing
        task-decorated functions, which the worker must import on startup. Will be
        called relative to current working directory. Can be set to the empty string to
        disable. Default ``'**/tasks.py'``
    """
    class Config:
        env_prefix = "FENNEL_"
        case_insensitive = True

    redis_url: str = "redis://127.0.0.1:6379"
    interface: str = "sync"
    processes: int = multiprocessing.cpu_count()
    concurrency: int = 8
    default_retries: int = 10
    retry_backoff: PyObject = "fennel.utils.backoff"
    read_timeout = 4000
    prefetch_count: int = 1
    heartbeat_timeout: float = 60
    heartbeat_interval: float = 6
    schedule_interval: float = 4
    maintenance_interval: float = 8
    task_timeout: int = 10
    grace_period: int = 30
    results_enabled: bool = True
    results_ttl: int = 60 * 60
    log_format: str = "console"
    log_level: str = "debug"
    autodiscover: str = "**/tasks.py"

    @validator("interface")
    def is_valid_interface(cls, value):
        assert value.lower() in ["sync", "async"]
        return value.lower()

    @validator("log_format")
    def is_valid_format(cls, value):
        assert value.lower() in ["console", "json"]
        return value.lower()

    @validator("autodiscover")
    def matches_python(cls, value):
        assert value == "" or value.endswith(".py")
        return value



"""
A collection of synchronous classes and functions to interact with the Fennel system.
"""

# flake8: noqa
from fennel.client.actions import purge_dead, read_dead, replay_dead
from fennel.client.results import AsyncResult
from fennel.client.task import Task
from fennel.client.waiting import gather, wait

__all__ = [
    "Task",
    "AsyncResult",
    "gather",
    "wait",
    "purge_dead",
    "read_dead",
    "replay_dead",
]



from typing import Any, Callable

from fennel.client.aio.actions import send
from fennel.client.aio.results import AsyncResult
from fennel.job import Job


class Task:
    def __init__(self, name: str, func: Callable, retries: int, app):
        self.name = name
        self.func = func
        self.max_retries = retries
        self.app = app

    async def delay(self, *args: Any, **kwargs: Any) -> AsyncResult:
        """
        Enqueue a task for execution by the workers.

        Similar to asyncio.create_task (but also works with non-async functions and runs
        on our Redis-backed task queue with distributed workers, automatic retry, and
        result storage with configurable TTL).

        The `args` and `kwargs` will be passed through to the task when executed.

        Examples
        --------
        >>> @app.task(retries=1)
        >>> async def foo(x, bar=None):
        ...     asyncio.sleep(x)
        ...     if bar == "mystr":
        ...         return False
        ...     return True
        ...
        >>> await foo.delay(1)
        >>> await foo.delay(2, bar="mystr")
        """
        job = Job(
            task=self.name,
            args=list(args),
            kwargs=kwargs,
            tries=0,
            max_retries=self.max_retries,
        )
        await send(self.app, job)
        return AsyncResult(job=job, app=self.app)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """
        Call the task-decorated function as a normal Python function. The fennel system
        will be completed bypassed.

        Examples
        --------
        >>> @app.task
        >>> def foo(x):
        ...     return x
        ...
        >>> foo(7)
        7
        """
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return f"Task(name={self.name})"



"""
Jobs have a number of statuses through their lifecycle. This module contains the
constants. If you have enqueued a task for execution, then you can obtain its status
as follows:

>>> x = mytask.delay()
>>> x.status()
EXECUTING
"""

#: The job's status is not stored in Redis. Presumably no action has been taken on the
#: job.
UNKNOWN = "UNKNOWN"

#: The job has been sent to Redis, but execution has not yet started.
SENT = "SENT"

#: A worker has received the job from the queue and has begun executing it.
EXECUTING = "EXECUTING"

#: Execution was successful and the job's result is ready (if results storage is
#: enabled).
SUCCESS = "SUCCESS"

#: Execution was not successful (an exception was raised) and a retry is scheduled
#: to occur in the future.
RETRY = "RETRY"

#: Execution was not successful (an exception was raised) and retries have been
#: exhausted, so the job is now in the dead-letter queue where it will remain
#: until manual intervention (via the CLI or client code).
DEAD = "DEAD"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fennel', 'fennel.client', 'fennel.client.aio', 'fennel.worker']

package_data = \
{'': ['*'], 'fennel.worker': ['lua/*']}

install_requires = \
['anyio>=2.0.0-beta.2,<3.0.0',
 'aredis>=1.1.8,<2.0.0',
 'click>=7.0,<8.0',
 'colorama>=0.4.1,<0.5.0',
 'hiredis>=1.1.0,<2.0.0',
 'pydantic>=1.5,<2.0',
 'redis>=3.3,<4.0',
 'structlog>=20.0,<21.0',
 'uvloop>=0.14.0,<0.15.0']

entry_points = \
{'console_scripts': ['fennel = fennel.cli:cli']}

setup_kwargs = {
    'name': 'fennel',
    'version': '0.3.0',
    'description': 'A task queue for Python based on Redis Streams.',
    'long_description': "## Fennel\n\nA task queue for Python 3.7+ based on Redis Streams with a Celery-like API.\n\nhttps://fennel.dev/\n\n| Note: This is an *alpha* release. The project is under development, breaking changes are likely. |\n| --- |\n\n### Features\n\n* Supports both sync (e.g. Django, Flask) and async (e.g. Starlette, FastAPI) code.\n* Sane defaults: at least once processing semantics, tasks acknowledged on completion.\n* Automatic retries with exponential backoff for fire-and-forget jobs.\n* Clear task statuses available (e.g. sent, executing, success).\n* Automatic task discovery (defaults to using ``**/tasks.py``).\n* Exceptionally small and understandable codebase.\n\n### Installation\n\n```bash\npip install fennel\n```\n\n### Basic Usage\n\nRun [Redis](https://redis.io) and then execute your code in `tasks.py`:\n```python\nfrom fennel import App\n\napp = App(name='myapp', redis_url='redis://127.0.0.1')\n\n\n@app.task\ndef foo(n):\n    return n\n\n\n# Enqueue a task to be executed in the background by a fennel worker process.\nfoo.delay(7)\n```\n\nMeanwhile, run the worker:\n```bash\n$ fennel worker --app tasks:app\n```\n\n### Asynchronous API\n\nFennel also supports an async API. If your code is running in an event loop\n(e.g. via [Starlette](https://www.starlette.io/) or\n[FastAPI](https://fastapi.tiangolo.com/)), you will want to use the async\ninterface instead:\n```python\nfrom fennel import App\n\napp = App(name='myapp', redis_url='redis://127.0.0.1', interface='async')\n\n\n@app.task\nasync def bar(x):\n    return x\n\n\nawait bar.delay(5)\n```\n\n### See also\n\nIf you need to ensure that all tasks for a given key are processed in-order,\nplease see our sister project [Runnel](https://github.com/mjwestcott/runnel).\n",
    'author': 'Matt Westcott',
    'author_email': 'm.westcott@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://fennel.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Tools for interacting with Fennel worker processes.
"""

# flake8: noqa
from fennel.worker.broker import Broker
from fennel.worker.discovery import autodiscover
from fennel.worker.executor import EXIT_COMPLETE, EXIT_SIGNAL, Executor
from fennel.worker.worker import start

__all__ = [
    "Broker",
    "autodiscover",
    "EXIT_COMPLETE",
    "EXIT_SIGNAL",
    "Executor",
    "start",
]
